/*
  httpclient.h - HTTP Client Library for Linux/Windows

  Copyright (c) 2020-2023 Sasapea's Lab. All right reserved.

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General
  Public License along with this library; if not, write to the
  Free Software Foundation, Inc., 59 Temple Place, Suite 330,
  Boston, MA  02111-1307  USA
*/
#ifndef __HTTPCLIENT_H
#define __HTTPCLIENT_H

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <string>
#include <vector>
#include <sstream>

#define HTTP_CODE_OK 200
#define HTTP_TIMEOUT 1000

typedef struct
{
  std::string name;
  std::string value;
} http_name_value_t;

class HttpHeader
{
  protected:

    std::vector<http_name_value_t> _headers;

    void addNameValue(std::vector<http_name_value_t> &vec, const char *name, int value)
    {
      addNameValue(vec, name, std::to_string(value).c_str());
    }

    void addNameValue(std::vector<http_name_value_t> &vec, const char *name, const char *value)
    {
      for(auto it = vec.begin(); it != vec.end(); ++it)
      {
        if (strcasecmp(it->name.c_str(), name) == 0)
        {
          it->value = value;
          return;
        }
      }
      vec.push_back({name, value});
    }

    const char *getNameValue(std::vector<http_name_value_t> &vec, const char *name, const char *defval = "")
    {
      for(auto it = vec.begin(); it != vec.end(); ++it)
      {
        if (strcasecmp(it->name.c_str(), name) == 0)
          return it->value.c_str();
      }
      return defval;
    }

    std::string listNameValue(std::vector<http_name_value_t> &vec, const char *sep, const char *dlm, bool last)
    {
      std::string list;
      for(auto it = vec.begin(); it != vec.end(); ++it)
      {
        if (list.size())
          list += dlm;
        list += it->name;
        list += sep;
        list += it->value;
      }
      if (last && list.size())
        list += dlm;
      return list;
    }

  public:

    void addHeader(const char *name, const char *value)
    {
      addNameValue(_headers, name, value);
    }

    void addHeader(const char *name, int value)
    {
      addNameValue(_headers, name, value);
    }

    const char *getHeader(const char *name, const char *defval = "")
    {
      return getNameValue(_headers, name, defval);
    }

    static const char *HTTP_METHOD_GET;
    static const char *HTTP_METHOD_POST;
    static const char *HTTP_HOST;
    static const char *HTTP_CONNECTION;
    static const char *HTTP_CONTENT_TYPE;
    static const char *HTTP_CONTENT_LENGTH;
    static const char *HTTP_MIME_TEXT_PLAIN;
    static const char *HTTP_MIME_X_WWW_FORM_URLENCODED;
    static const char *HTTP_LF;
};

const char *HttpHeader::HTTP_METHOD_GET                 = "GET";
const char *HttpHeader::HTTP_METHOD_POST                = "POST";
const char *HttpHeader::HTTP_HOST                       = "Host";
const char *HttpHeader::HTTP_CONNECTION                 = "Connection";
const char *HttpHeader::HTTP_CONTENT_TYPE               = "Content-Type";
const char *HttpHeader::HTTP_CONTENT_LENGTH             = "Content-Length";
const char *HttpHeader::HTTP_MIME_TEXT_PLAIN            = "text/plain";
const char *HttpHeader::HTTP_MIME_X_WWW_FORM_URLENCODED = "application/x-www-form-urlencoded";
const char *HttpHeader::HTTP_LF                         = "\r\n";

class HttpRequest : public HttpHeader
{
  private:

    std::vector<http_name_value_t> _queries;
    std::string _request;

  public:

    void clear(void)
    {
      _headers.clear();
      _queries.clear();
      addHeader(HTTP_CONNECTION, "close");
    }

    void addQuery(const char *name, const char *value)
    {
      addNameValue(_queries, name, value);
    }

    const char *getHttpString(const char *host, const char *path, const char *method, const char *content = "")
    {
      std::vector<http_name_value_t> headers = _headers;
      std::string queries = listNameValue(_queries, "=", "&", false);
      std::string type = getNameValue(headers, HTTP_CONTENT_TYPE, "");
      std::string data = content;
      if (type.size() == 0)
        type = ((data.size() == 0) && queries.size() ? HTTP_MIME_X_WWW_FORM_URLENCODED : HTTP_MIME_TEXT_PLAIN);
      if (type == HTTP_MIME_X_WWW_FORM_URLENCODED)
        data = queries;
      _request = method;
      _request += " /";
      _request += path;
      if (strcasecmp(type.c_str(), HTTP_MIME_X_WWW_FORM_URLENCODED) && queries.size())
      {
        _request += '?';
        _request += queries;
      }
      _request += " HTTP/";
      if (host && *host)
      {
        _request += "1.1";
        addNameValue(headers, HTTP_HOST, host);
      }
      else
        _request += "1.0";
      _request += HTTP_LF;
      addNameValue(headers, HTTP_CONTENT_TYPE  , type.c_str());
      addNameValue(headers, HTTP_CONTENT_LENGTH, data.size());
      _request += listNameValue(headers, ": ", HTTP_LF, true);
      _request += HTTP_LF;
      _request += data;
      return _request.c_str();
    }
};

class HttpResponse : public HttpHeader
{
  private:

    std::string _content;
    std::string _result[3];

  public:

    void clear(void)
    {
      _headers.clear();
      _content.clear();
      _result[0] = "";
      _result[1] = "-1";
      _result[2] = "";
    }

    void parse(const char *response)
    {
      std::string issstr = response;
      std::istringstream iss(issstr);
      std::string line;
      clear();
      while (std::getline(iss, line))
      {
        int len = line.length();
        if ((len > 0) && (line[len - 1] == '\r'))
          line.resize(--len);
        if (len == 0)
          break;
        if (line.find("HTTP/") == 0)
        {
          int cnt = 0;
          int pos = 5;
          while (pos < len)
          {
            int sep = pos;
            while ((pos < len) && (line[pos] != ' '))
             ++pos;
            _result[cnt++] = line.substr(sep, pos - sep);
            while ((pos < len) && (line[pos] == ' '))
              ++pos;
            if (cnt == sizeof(_result) / sizeof(_result[0]) - 1)
            {
              _result[cnt] = line.substr(pos);
              break;
            }
          }
        }
        else
        {
          int pos = 0;
          while ((pos < len) && (line[pos] != ':'))
           ++pos;
          int sep = pos++;
          while ((pos < len) && (line[pos] == ' '))
            ++pos;
          addHeader(line.substr(0, sep).c_str(), line.substr(pos).c_str());
        }
      }
      _content = issstr.substr((int)iss.tellg());
    }

    const char *getHttpVersion(void)
    {
      return _result[0].c_str();
    }

    int getHttpStatus(void)
    {
      return std::stoi(_result[1]);
    }

    const char *getHttpMessage(void)
    {
      return _result[2].c_str();
    }

    const char *getContentType(void)
    {
      return getHeader(HTTP_CONTENT_TYPE, "");
    }

    int getContentLength(void)
    {
      return std::stoi(getHeader(HTTP_CONTENT_LENGTH, "-1"), nullptr, 10);
    }

    const char *content(void)
    {
      return _content.c_str();
    }

    std::string toString()
    {
      std::string str = listNameValue(_headers, ": ", "\n", true);
      str += "\n";
      str += _content.c_str();
      return str;
    }
};

#if defined(_MSC_VER) || defined(__MINGW32__) || defined(__MINGW64__)
  #define __WINSOCK__
#endif

// socket
#ifdef __WINSOCK__
  #include <ws2tcpip.h>
  #define MSG_NOSIGNAL 0
  #define ssize_t int
#else
  #include <time.h>
  #include <unistd.h>
  #include <sys/types.h>
  #include <sys/socket.h>
  #include <netinet/in.h>
  #include <arpa/inet.h>
  #define closesocket(s)  close(s)
#endif

class HttpClient : public HttpRequest
{
  public:

    int request(HttpResponse &response, const char *host, int port, const char *path, const char *method, const char *content = "")
    {
      struct sockaddr_in addr;
      memset(&addr, 0, sizeof(addr));
      addr.sin_family = AF_INET;
      addr.sin_port   = htons(port);
      addr.sin_addr.s_addr = inet_addr(host);
      response.clear();
      int sock = socket(PF_INET, SOCK_STREAM, IPPROTO_TCP);
      if (sock != -1)
      {
        if (connect(sock, (struct sockaddr *)&addr, sizeof(addr)) != -1)
        {
          const char *http = getHttpString(host, path, method, content);
          ssize_t len = strlen(http);
          if (send(sock, http, len, MSG_NOSIGNAL) == len)
          {
            int len, cnt = 0;
            char *buf = (char *)malloc(len = 1024);
            if (buf)
            {
              struct timeval tov = { HTTP_TIMEOUT / 1000, (HTTP_TIMEOUT % 1000) * 1000 };
              while (1)
              {
                char tmp[256];
                fd_set rdfds;
                int n;
                FD_ZERO(&rdfds);
                FD_SET(sock, &rdfds);
                if ((n = select(sock + 1, &rdfds, NULL, NULL, &tov)) < 0)
                  break;
                if (n > 0)
                {
                  if ((n = recv(sock, tmp, sizeof(tmp), 0)) < 0)
                    break;
                }
                if (n == 0)
                {
                  buf[cnt] = 0;
                  response.parse(buf);
                  break;
                }
                if (cnt + n >= len)
                {
                  len += 1024;
                  char *p = (char *)realloc(buf, len);
                  if (!p)
                    break;
                  buf = p;
                }
                memcpy(buf + cnt, tmp, n);
                cnt += n;
                tov.tv_sec  = 0;
                tov.tv_usec = 20000; // 20ms
              }
              free(buf);
            }
          }
        }
        closesocket(sock);
      }
      return response.getHttpStatus();
    }

    static bool startup(void)
    {
#ifdef __WINSOCK__
      WSADATA wsaData;
      return WSAStartup(MAKEWORD(2, 2), &wsaData) == 0;
#else
      return true;
#endif
    }

    static void cleanup(void)
    {
#ifdef __WINSOCK__
      WSACleanup();
#endif
    }
};

#endif
