/*
  tapo.h - TP-Link TAPO Control Library for P100/P105/P110

  Copyright (c) 2020-2023 Sasapea's Lab. All right reserved.

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General
  Public License along with this library; if not, write to the
  Free Software Foundation, Inc., 59 Temple Place, Suite 330,
  Boston, MA  02111-1307  USA
*/
#ifndef __TAPO_H
#define __TAPO_H

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <stdint.h>
#include <stdbool.h>
#include <string.h>
#include <time.h>
#include <string>

#include "httpclient.h"

#include "libb64/cencode.h"
#include "libb64/cdecode.h"

// Requires install. (apt install mbedtls-dev)
#include "mbedtls/entropy.h"
#include "mbedtls/ctr_drbg.h"
#include "mbedtls/cipher.h"
#include "mbedtls/pk.h"

// Requires download. (https://github.com/nlohmann/json)
#include "nlohmann/json.hpp"
using json = nlohmann::json;  // for convenience

typedef enum
{
  TAPO_ERROR_SUCCESS                       = 0,
  TAPO_ERROR_INCORRECT_REQUEST             = -1002,
  TAPO_ERROR_JSON_FORMATTING_ERROR         = -1003,
  TAPO_ERROR_INVALID_PUBLIC_KEY_LENGTH     = -1010, // invalid cookie
  TAPO_ERROR_INVALID_TERMINALUUID          = -1012,
  TAPO_ERROR_INVALID_REQUEST_OR_CREDENTIAL = -1501,
  TAPO_ERROR_DEVICE_BUSY                   =  9999,
} TAPO_ERROR;

#define TAPO_SESSION_TIMEOUT(t) ((t) * 60 * 1000)

#define TAPO_MD_LENGTH       20
#define TAPO_CIPHER_KEY_BITS 128
#define TAPO_RSA_KEY_BITS    1024
#define TAPO_RSA_EXPONENT    65537
#define TAPO_SEED_PERSONAL   "rsa_genkey"

#define TAPO_METHOD_HANDSHAKE         "handshake"
#define TAPO_METHOD_LOGIN_DEVICE      "login_device"
#define TAPO_METHOD_GET_DEVICE_INFO   "get_device_info"
#define TAPO_METHOD_SET_DEVICE_INFO   "set_device_info"
#define TAPO_METHOD_GET_ENERGY_USAGE  "get_energy_usage"
#define TAPO_METHOD_SECUREPASSTHROUGH "securePassthrough"

#define TAPO_DEBUG_PRINTF(format, ...)   debug_printf(__FILE__, __LINE__, format, __VA_ARGS__)
#define TAPO_DEBUG_MBEDTLS(name, status) mbedtls_status(__FILE__, __LINE__, name, status)

class TAPO
{
  private:

    std::string  _hostname;
    std::string  _username;
    std::string  _password;
    HttpClient   _httpClient;
    HttpResponse _httpResponse;
    int          _httpStatus;
    int          _tapoStatus;
    int          _mbedStatus;
    std::string  _tapoCookie;
    std::string  _tapoToken;
    std::string  _terminalUUID;
    uint32_t     _sessionTimeout;
    uint32_t     _sessionStart;
    std::string  _publicKey;
    std::string  _privateKey;
    uint8_t      _cipherKey[TAPO_CIPHER_KEY_BITS >> 3];
    uint8_t      _cipherIv [TAPO_CIPHER_KEY_BITS >> 3];
    uint8_t      _buffer[2048];

    void debug_printf(const char *file, int line, const char *format, ...)
    {
      printf("[%s:%d] ", file, line);
      va_list ap;
      va_start(ap, format);
      vprintf(format, ap);
      va_end(ap);
    }

    int mbedtls_status(const char *file, int line, const char *name, int status)
    {
      _mbedStatus = status;
      if (status)
        debug_printf(file, line, "%s = -0x%04X\n", name, abs(status));
      return status;
    }

    std::string toHEX(uint8_t *buf, uint16_t len)
    {
      static const char HEX_DIGITS[] = "0123456789abcdef";
      std::string hex;
      while (len-- > 0)
      {
        uint8_t b = *buf++;
        hex += HEX_DIGITS[b >> 4];
        hex += HEX_DIGITS[b & 0x0F];
      }
      return hex;
    }

    int b64encode(const uint8_t *bin, int len, uint8_t *b64)
    {
      base64_encodestate state;
      base64_init_encodestate(&state);
      len =  base64_encode_block((const char *)bin, len, (char *)b64, &state);
      return len + base64_encode_blockend((char *)b64 + len, &state);
    }

    int b64decode(const uint8_t *b64, int len, uint8_t *bin)
    {
      base64_decodestate state;
      base64_init_decodestate(&state);
      return base64_decode_block((const char *)b64, len, (char *)bin, &state);
    }

    void replaceAll(std::string &replacedStr, std::string from, std::string to)
    {
      if (!from.empty())
      {
        int pos = 0;
        while ((pos = replacedStr.find(from, pos)) != (int)std::string::npos)
        {
          replacedStr.replace(pos, from.length(), to);
          pos += to.length();
        }
      }
    }

    std::string removePemLF(const std::string &pem)
    {
      int    head = pem.find_first_of('\n');
      int    tail = pem.find_last_of('\n', pem.length() - 2);
      std::string body = pem.substr(head + 1, tail - (head + 1));
      replaceAll(body, std::string("\n"), std::string(""));
      return pem.substr(0, head + 1) + body + pem.substr(tail + 1);
    }

    bool generateKeyPair(void)
    {
      bool rv = false;
      int olen;
      mbedtls_pk_context       pk;
      mbedtls_entropy_context  entropy;
      mbedtls_ctr_drbg_context ctr_drbg;
      _buffer[0] = 0;
      mbedtls_pk_init(&pk);
      mbedtls_entropy_init(&entropy);
      mbedtls_ctr_drbg_init(&ctr_drbg);
      if (TAPO_DEBUG_MBEDTLS("mbedtls_ctr_drbg_seed()", mbedtls_ctr_drbg_seed(&ctr_drbg, mbedtls_entropy_func, &entropy, (uint8_t *)TAPO_SEED_PERSONAL, strlen(TAPO_SEED_PERSONAL)))) ;
      else if (TAPO_DEBUG_MBEDTLS("mbedtls_pk_setup()", mbedtls_pk_setup(&pk, mbedtls_pk_info_from_type(MBEDTLS_PK_RSA)))) ;
      else if (TAPO_DEBUG_MBEDTLS("mbedtls_rsa_gen_key()", mbedtls_rsa_gen_key(mbedtls_pk_rsa(pk), mbedtls_ctr_drbg_random, &ctr_drbg, TAPO_RSA_KEY_BITS, TAPO_RSA_EXPONENT))) ;
      else if (TAPO_DEBUG_MBEDTLS("mbedtls_pk_write_pubkey_pem()", mbedtls_pk_write_pubkey_pem(&pk, _buffer, sizeof(_buffer)))) ;
      else if ((olen = strlen((char *)_buffer)) == 0) TAPO_DEBUG_PRINTF("invalid public key length = %d\n", olen);
      else if (TAPO_DEBUG_MBEDTLS("mbedtls_pk_write_key_pem()", mbedtls_pk_write_key_pem(&pk, _buffer + olen + 1, sizeof(_buffer) - olen - 1))) ;
      else
      {
        _publicKey  = removePemLF(std::string((char *)_buffer));
        _privateKey = (char *)(_buffer + olen + 1);
        rv = true;
      }
      mbedtls_ctr_drbg_free(&ctr_drbg);
      mbedtls_entropy_free(&entropy);
      mbedtls_pk_free(&pk);
      return rv;
    }

    bool cipherKey(const std::string &key)
    {
      bool rv = false;
      mbedtls_pk_context       pk;
      mbedtls_entropy_context  entropy;
      mbedtls_ctr_drbg_context ctr_drbg;
      uint16_t ilen = key.length();
      size_t   olen = b64decode((uint8_t *)key.c_str(), ilen, _buffer + ilen);
      mbedtls_pk_init(&pk);
      mbedtls_entropy_init(&entropy);
      mbedtls_ctr_drbg_init(&ctr_drbg);
      if (TAPO_DEBUG_MBEDTLS("mbedtls_ctr_drbg_seed()", mbedtls_ctr_drbg_seed(&ctr_drbg, mbedtls_entropy_func, &entropy, (uint8_t *)TAPO_SEED_PERSONAL, strlen(TAPO_SEED_PERSONAL)))) ;
      else if (TAPO_DEBUG_MBEDTLS("mbedtls_pk_parse_key()", mbedtls_pk_parse_key(&pk, (uint8_t *)_privateKey.c_str(), _privateKey.length() + 1, NULL, 0))) ;
      else if (TAPO_DEBUG_MBEDTLS("mbedtls_pk_decrypt()", mbedtls_pk_decrypt(&pk, _buffer + ilen, olen, _buffer, &olen, ilen, mbedtls_ctr_drbg_random, &ctr_drbg))) ;
      else if (olen != sizeof(_cipherKey) + sizeof(_cipherIv)) TAPO_DEBUG_PRINTF("invalid cipher key length = %d\n", olen);
      else
      {
        memcpy(_cipherKey, _buffer, sizeof(_cipherKey));
        memcpy(_cipherIv, _buffer + sizeof(_cipherKey), sizeof(_cipherIv));
        rv = true;
      }
      mbedtls_ctr_drbg_free(&ctr_drbg);
      mbedtls_entropy_free(&entropy);
      mbedtls_pk_free(&pk);
      return rv;
    }

    std::string digest(const char *msg)
    {
      return TAPO_DEBUG_MBEDTLS("mbedtls_md(mbedtls_md_info_from_type()", mbedtls_md(mbedtls_md_info_from_type(MBEDTLS_MD_SHA1), (const uint8_t *)msg, strlen(msg), _buffer))
        ? std::string() : toHEX(_buffer, TAPO_MD_LENGTH);
    }

    size_t cipher0(mbedtls_operation_t operation, const uint8_t *input, size_t ilen, uint8_t *output)
    {
      size_t olen, olen2;
      mbedtls_cipher_context_t cipher;
      mbedtls_cipher_init(&cipher);
      if (TAPO_DEBUG_MBEDTLS("mbedtls_cipher_setup()", mbedtls_cipher_setup(&cipher, mbedtls_cipher_info_from_values(MBEDTLS_CIPHER_ID_AES, TAPO_CIPHER_KEY_BITS, MBEDTLS_MODE_CBC)))) ;
      else if (TAPO_DEBUG_MBEDTLS("mbedtls_cipher_set_padding_mode()", mbedtls_cipher_set_padding_mode(&cipher, MBEDTLS_PADDING_PKCS7))) ;
      else if (TAPO_DEBUG_MBEDTLS("mbedtls_cipher_set_iv()", mbedtls_cipher_set_iv(&cipher, _cipherIv, sizeof(_cipherIv)))) ;
      else if (TAPO_DEBUG_MBEDTLS("mbedtls_cipher_setkey()", mbedtls_cipher_setkey(&cipher, _cipherKey, TAPO_CIPHER_KEY_BITS, operation))) ;
      else if (TAPO_DEBUG_MBEDTLS("mbedtls_cipher_update()", mbedtls_cipher_update(&cipher, input, ilen, output, &olen))) ;
      else if (TAPO_DEBUG_MBEDTLS("mbedtls_cipher_finish()", mbedtls_cipher_finish(&cipher, output + olen, &olen2)))  olen = olen2 = 0;
      mbedtls_cipher_free(&cipher);
      return _mbedStatus ? 0 : olen + olen2;
    }

    std::string encrypt(const std::string &plain)
    {
      size_t ilen = plain.length();
      size_t olen = cipher0(MBEDTLS_ENCRYPT, (uint8_t *)plain.c_str(), ilen, _buffer + ilen);
      _buffer[b64encode(_buffer + ilen, olen, _buffer)] = 0;
      return std::string((char *)_buffer);
    }

    std::string decrypt(const std::string &cipher)
    {
      size_t ilen = cipher.length();
      size_t olen = b64decode((uint8_t *)cipher.c_str(), ilen, _buffer + ilen);
      _buffer[cipher0(MBEDTLS_DECRYPT, _buffer + ilen, olen, _buffer)] = 0;
      return std::string((char *)_buffer);
    }

    bool request0(const std::string &payload)
    {
      _httpStatus = 0;
      _tapoStatus = 0;
      _mbedStatus = 0;
      _httpClient.clear();
      if (_tapoToken.length())
        _httpClient.addQuery("token", _tapoToken.c_str());
      if (_tapoCookie.length())
        _httpClient.addHeader("Cookie", _tapoCookie.c_str());
      _httpClient.addHeader("Content-Type", "application/json;charset=UTF-8");
      _httpStatus = _httpClient.request(_httpResponse, _hostname.c_str(), 80, "app", "POST", payload.c_str());
      if (_httpStatus != HTTP_CODE_OK)
        return false;
      std::string cookie = _httpResponse.getHeader("Set-Cookie", "");
      if (cookie.length())
      {
        _sessionStart = millis();
        // parse cookie
        int left  = 0;
        int right = -1;
        while (right + 1 < (int)cookie.length())
        {
          left  = right + 1;
          right = cookie.find_first_of(';', left);
          if (right < 0)
            right = cookie.length();
          int mid = cookie.find_first_of('=', left);
          if (mid < 0)
            mid = right;
          std::string name  = cookie.substr(left, mid - left);
          std::string value = cookie.substr(mid + 1, right - (mid + 1));
          if (strcasecmp(name.c_str(), "TP_SESSIONID") == 0)
            _tapoCookie = cookie.substr(left, right - left);
          if (strcasecmp(name.c_str(), "TIMEOUT") == 0)
            _sessionTimeout = TAPO_SESSION_TIMEOUT(atoi(value.c_str()) - 1);
        }
      }
      return true;
    }

    const char *request(const std::string &payload)
    {
      int retry = 3;
      while (1)
      {
        if (request0(payload))
          return _httpResponse.content();
        if (--retry == 0)
          break;
        delay(1000);
      }
      return "";
    }

    bool handshake(void)
    {
      json payload;
      payload["method"] = TAPO_METHOD_HANDSHAKE;
      payload["params"]["key"] = _publicKey;
      payload["requestTimeMils"] = millis();
      std::string response = request(payload.dump());
      if (response.length())
      {
        json res = json::parse(response);
        _tapoStatus = res["error_code"].get<int>();
        if (_tapoStatus == 0)
        {
          if (cipherKey(res["result"]["key"].get<std::string>()) && _tapoCookie.length())
            return true;
        }
      }
      return false;
    }

    bool login(void)
    {
      json payload;
      payload["method"] = TAPO_METHOD_LOGIN_DEVICE;
      payload["params"]["username"] = _username;
      payload["params"]["password"] = _password;
      payload["requestTimeMils"] = millis();
      json envelope;
      envelope["method"] = TAPO_METHOD_SECUREPASSTHROUGH;
      envelope["params"]["request"] = encrypt(payload.dump());
      std::string response = request(envelope.dump());
      if (response.length())
      {
        json res = json::parse(response);
        _tapoStatus = res["error_code"].get<int>();
        if (_tapoStatus == 0)
        {
          res = json::parse(decrypt(res["result"]["response"].dump()));
          _tapoStatus = res["error_code"].get<int>();
          if (_tapoStatus == 0)
          {
            _tapoToken = res["result"]["token"].get<std::string>();
            if (_tapoToken.length())
              return true;
          }
        }
      }
      return false;
    }

    bool beginSession(void)
    {
      if (_tapoCookie.length() && _tapoToken.length())
      {
        if (millis() - _sessionStart < _sessionTimeout)
          return true;
        endSession();
      }
      return handshake() && login();
    }

    void endSession(void)
    {
      _tapoCookie.clear();
      _tapoToken.clear();
      _terminalUUID.clear();
    }

    std::string getTerminalUUID(void)
    {
      if (_terminalUUID.length() == 0)
      {
        std::string info;
        if (getDeviceInfo(info))
          _terminalUUID = json::parse(info)["result"]["mac"].get<std::string>();
      }
      return _terminalUUID;
    }

  public:

    TAPO(void)
    : _httpStatus(0)
    , _tapoStatus(0)
    , _mbedStatus(0)
    , _sessionTimeout(TAPO_SESSION_TIMEOUT(1440 - 1))
    {
    }

    virtual ~TAPO(void)
    {
    }

    bool begin(const char *hostname, const char *username, const char *password)
    {
      // support for winsock
      if (!HttpClient::startup())
      {
        TAPO_DEBUG_PRINTF("HttpClient::startup() failed.\n", 0);
        return false;
      }
      if (!generateKeyPair())
      {
        TAPO_DEBUG_PRINTF("generateKeyPair() failed.\n", 0);
        return false;
      }
      // hostname (ipaddress only)
      _hostname = hostname;
      // username
      std::string id = digest(username);
      _buffer[b64encode((uint8_t *)id.c_str(), id.length(), _buffer)] = 0;
      _username = (char *)_buffer;
      // password
      _buffer[b64encode((uint8_t *)password, strlen(password), _buffer)] = 0;
      _password = (char *)_buffer;
      //
      endSession();
      return true;
    }

    void end(void)
    {
      // support for winsock
      HttpClient::cleanup();
    }

    int httpStatus(void)
    {
      return _httpStatus;
    }

    int tapoStatus(void)
    {
      return _tapoStatus;
    }

    int mbedStatus(void)
    {
      return _mbedStatus;
    }

    /*
     * std::string info;
     * getDeviceInfo(info);
     * decode(json::parse(info)["result"]["nickname" or "ssid"].get<std::string>());
     */
    std::string decode(const std::string &str)
    {
      _buffer[b64decode((uint8_t *)str.c_str(), str.length(), _buffer)] = 0;
      return std::string((char *)_buffer);
    }

    bool getEnergyUsage(std::string &info)
    {
      if (beginSession())
      {
        json payload;
        payload["method"] = TAPO_METHOD_GET_ENERGY_USAGE;
        payload["requestTimeMils"] = millis();
        json envelope;
        envelope["method"] = TAPO_METHOD_SECUREPASSTHROUGH;
        envelope["params"]["request"] = encrypt(payload.dump());
        std::string response = request(envelope.dump());
        if (response.length())
        {
          json res = json::parse(response);
          _tapoStatus = res["error_code"].get<int>();
          if (_tapoStatus == 0)
          {
            info = decrypt(res["result"]["response"].dump());
            return true;
          }
        }
      }
      endSession();
      return false;
    }

    bool getDeviceInfo(std::string &info)
    {
      if (beginSession())
      {
        json payload;
        payload["method"] = TAPO_METHOD_GET_DEVICE_INFO;
        payload["requestTimeMils"] = millis();
        json envelope;
        envelope["method"] = TAPO_METHOD_SECUREPASSTHROUGH;
        envelope["params"]["request"] = encrypt(payload.dump());
        std::string response = request(envelope.dump());
        if (response.length())
        {
          json res = json::parse(response);
          _tapoStatus = res["error_code"].get<int>();
          if (_tapoStatus == 0)
          {
            info = decrypt(res["result"]["response"].dump());
            return true;
          }
        }
      }
      endSession();
      return false;
    }

    bool setDeviceOn(bool on)
    {
      if (beginSession())
      {
        json payload;
        payload["method"] = TAPO_METHOD_SET_DEVICE_INFO;
        payload["params"]["device_on"] = on;
        payload["terminalUUID"] = getTerminalUUID();
        payload["requestTimeMils"] = millis();
        json envelope;
        envelope["method"] = TAPO_METHOD_SECUREPASSTHROUGH;
        envelope["params"]["request"] = encrypt(payload.dump());
        std::string response = request(envelope.dump());
        if (response.length())
        {
          json res = json::parse(response);
          _tapoStatus = res["error_code"].get<int>();
          if (_tapoStatus == 0)
            return true;
        }
      }
      endSession();
      return false;
    }

    static unsigned long micros(void)
    {
      struct timespec ts;

      clock_gettime(CLOCK_MONOTONIC, &ts);
      return (unsigned long)((unsigned long long)ts.tv_sec * 1000000) + (ts.tv_nsec / 1000);
    }

    static unsigned long millis(void)
    {
      struct timespec ts;

      clock_gettime(CLOCK_MONOTONIC, &ts);
      return (unsigned long)((unsigned long long)ts.tv_sec * 1000) + (ts.tv_nsec / 1000000);
    }

    static void delay(unsigned long ms)
    {
      struct timespec ts;

      clock_gettime(CLOCK_MONOTONIC, &ts);
      ts.tv_sec  += (ms / 1000);
      ts.tv_nsec += (ms % 1000) * 1000000;
      if (ts.tv_nsec >= 1000000000)
      {
        ts.tv_nsec -= 1000000000;
        ts.tv_sec++;
      }
      clock_nanosleep(CLOCK_MONOTONIC, TIMER_ABSTIME, &ts, NULL);
    }

    static void delayMicroseconds(unsigned int us)
    {
      struct timespec ts;

      clock_gettime(CLOCK_MONOTONIC, &ts);
      ts.tv_sec  += (us / 1000000);
      ts.tv_nsec += (us % 1000000) * 1000;
      if (ts.tv_nsec >= 1000000000)
      {
        ts.tv_nsec -= 1000000000;
        ts.tv_sec++;
      }
      clock_nanosleep(CLOCK_MONOTONIC, TIMER_ABSTIME, &ts, NULL);
    }
};

#endif
